use std::path::{Path, PathBuf};
use std::fs::File;
use std::io;

use meshtastic::protobufs::config::{BluetoothConfig, DeviceConfig, DisplayConfig, LoRaConfig, NetworkConfig, PayloadVariant, PositionConfig, PowerConfig};
use meshtastic::protobufs::module_config::{AmbientLightingConfig, AudioConfig, CannedMessageConfig, DetectionSensorConfig, ExternalNotificationConfig, MqttConfig, NeighborInfoConfig, PaxcounterConfig, RangeTestConfig, RemoteHardwareConfig, SerialConfig, StoreForwardConfig, TelemetryConfig};
use rusqlite::{Connection, Result};
use rusqlite_migration::{Migrations, M};
use serde::{Deserialize, Serialize};
use tracing::{debug, error, info};
use lazy_static::lazy_static;
use meshtastic::protobufs::{Channel, DeviceMetrics, MyNodeInfo, NodeInfo, Position, User};

use crate::CONFIG;

pub(crate) struct Database {
    conn: Connection,
}

// Migrations
lazy_static! {
    static ref MIGRATIONS: Migrations<'static> = Migrations::new(vec![
        // Define your migrations here
        M::up(include_str!("../db/migrations/0001_initial.sql")),
    ]);
}

fn u32_vec_to_u8_vec(vec: &Vec<u32>) -> Vec<u8> {
    vec.iter()
        .flat_map(|x| x.to_be_bytes().to_vec())
        .collect()
}

#[derive(Serialize, Deserialize)]
struct RemoteHardwarePin {
    gpio_pin: u32,
    name: String,
    r#type: i32,
}

/// This contains all the code for reading and writing from the database.
/// Holy shit levels of boilerplate.
/// This is done like this instead of just saving BLOBs because I really want to use SQL itself to quickly query data.
/// Any parsing/indexing/long term statistics code I write isn't going to be even close to the level of optimisation
/// that rusqlite has.
impl Database {
    /// Opens the database, running migrations and creating if it doesn't exist.
    pub fn open() -> Result<Self> {
        let db_path = PathBuf::from(&CONFIG.database.db_name);
        // Pre-check, does the file exist?
        if !db_path.exists() {
            // If not, create the file.
            match File::create(&db_path) {
                Ok(_) => info!("Created database stub file."),
                Err(e) => error!("Failed to create file: {:?}", e),
            }
        }

        // Open it up, and run migrations!
        let mut conn = Connection::open(&db_path)?;
        // Apply migrations
        MIGRATIONS.to_latest(&mut conn).expect("Migration failed");
        Ok(Database { conn })
    }

    /// Updates the database with a new MyNodeInfo packet
    pub fn update_mynodeinfo(my_node_info: &MyNodeInfo) -> Result<bool> {
        // Open DB
        let db = Database::open().expect("Failure to open database");

        db.conn.execute(
            "INSERT INTO MyNodeInfo (my_node_num, reboot_count, min_app_version) VALUES (?1, ?2, ?3)",
            rusqlite::params![my_node_info.my_node_num, my_node_info.reboot_count, my_node_info.min_app_version],
        )?;
        debug!("Saved MyNodeInfo to database");
        Ok(true)
    }

    /// Updates the database with a new NodeInfo packet
    pub fn update_nodeinfo(node_info: &NodeInfo) -> Result<bool> {
        // Open DB
        let db = Database::open().expect("Failure to open database");

        db.conn.execute(
            "INSERT INTO Node (
                num, user_id, user_lname, user_sname, hw_model, is_licensed, role, 
                latitude_i, longitude_i, altitude, location_source, gps_timestamp, 
                altitude_source, pdop, hdop, vdop, gps_accuracy, ground_speed, 
                ground_track, fix_quality, fix_type, sats_in_view, sensor_id, 
                next_update, seq_number, precision_bits, snr, last_heard, 
                battery_level, voltage, channel_utilization, air_util_tx, 
                channel, via_mqtt, hops_away
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, 
                ?15, ?16, ?17, ?18, ?19, ?20, ?21, ?22, ?23, ?24, ?25, ?26, 
                ?27, ?28, ?29, ?30, ?31, ?32, ?33, ?34, ?35
            )",
            rusqlite::params![
                node_info.num,
                // This is gibberish to me, so lets explain.
                // We're mapping the value user.id as the value, but chucking a NULL in if it's a None type.
                node_info.user.as_ref().map(|user| user.id.clone()),
                node_info.user.as_ref().map(|user| user.long_name.clone()),
                node_info.user.as_ref().map(|user| user.short_name.clone()),
                node_info.user.as_ref().map(|user| user.hw_model),
                node_info.user.as_ref().map(|user| user.is_licensed),
                node_info.user.as_ref().map(|user| user.role),
                node_info.position.as_ref().map(|position| position.latitude_i),
                node_info.position.as_ref().map(|position| position.longitude_i),
                node_info.position.as_ref().map(|position| position.altitude),
                node_info.position.as_ref().map(|position| position.location_source),
                node_info.position.as_ref().map(|position| position.timestamp),
                node_info.position.as_ref().map(|position| position.altitude_source),
                node_info.position.as_ref().map(|position| position.pdop),
                node_info.position.as_ref().map(|position| position.hdop),
                node_info.position.as_ref().map(|position| position.vdop),
                node_info.position.as_ref().map(|position| position.gps_accuracy),
                node_info.position.as_ref().map(|position| position.ground_speed),
                node_info.position.as_ref().map(|position| position.ground_track),
                node_info.position.as_ref().map(|position| position.fix_quality),
                node_info.position.as_ref().map(|position| position.fix_type),
                node_info.position.as_ref().map(|position| position.sats_in_view),
                node_info.position.as_ref().map(|position| position.sensor_id),
                node_info.position.as_ref().map(|position| position.next_update),
                node_info.position.as_ref().map(|position| position.seq_number),
                node_info.position.as_ref().map(|position| position.precision_bits),
                node_info.snr,
                node_info.last_heard,
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.battery_level),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.voltage),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.channel_utilization),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.air_util_tx),
                node_info.channel,
                node_info.via_mqtt,
                node_info.hops_away
            ]
        )?;

        debug!("Saved NodeInfo to database");
        Ok(true)
    }

    /// Updates the database with a new Channel packet
    pub fn update_channel(channel: &Channel) -> Result<bool> {
        // Open DB
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO Channel (
                channel_index, psk, name, id, uplink_enabled, downlink_enabled, position_precision, role
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8
            )",
            rusqlite::params![
                channel.index,
                channel.settings.as_ref().map(|settings| settings.psk.clone()),
                channel.settings.as_ref().map(|settings| settings.name.clone()),
                channel.settings.as_ref().map(|settings| settings.id),
                channel.settings.as_ref().map(|settings| settings.uplink_enabled),
                channel.settings.as_ref().map(|settings| settings.downlink_enabled),
                channel.settings.as_ref().and_then(|settings| settings.module_settings.as_ref()).map(|module_settings| module_settings.position_precision),
                channel.role
            ]
        )?;
        debug!("Saved Channel to database");
        Ok(true)
    }

    pub fn update_deviceconfig(config: &DeviceConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO DeviceConfig (
                role, serial_enabled, debug_log_enabled, button_gpio, buzzer_gpio, rebroadcast_mode,
                node_info_broadcast_secs, double_tap_as_button_press, is_managed, disable_triple_click
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10
            )",
            rusqlite::params![
                config.role,
                config.serial_enabled,
                config.debug_log_enabled,
                config.button_gpio,
                config.buzzer_gpio,
                config.rebroadcast_mode,
                config.node_info_broadcast_secs,
                config.double_tap_as_button_press,
                config.is_managed,
                config.disable_triple_click
            ]
        )?;
        debug!("Saved DeviceConfig to database");
        Ok(true)
    }

    pub fn update_position_config(config: &PositionConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO PositionConfig (
                position_broadcast_secs, position_broadcast_smart_enabled, fixed_position, gps_enabled, 
                gps_update_interval, gps_attempt_time, position_flags, rx_gpio, tx_gpio, 
                broadcast_smart_minimum_distance, broadcast_smart_minimum_interval_secs, gps_en_gpio, 
                gps_mode
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13
            )",
            rusqlite::params![
                config.position_broadcast_secs,
                config.position_broadcast_smart_enabled,
                config.fixed_position,
                config.gps_enabled,
                config.gps_update_interval,
                config.gps_attempt_time,
                config.position_flags,
                config.rx_gpio,
                config.tx_gpio,
                config.broadcast_smart_minimum_distance,
                config.broadcast_smart_minimum_interval_secs,
                config.gps_en_gpio,
                config.gps_mode
            ]
        )?;
        debug!("Saved PositionConfig to database");
        Ok(true)
    }

    pub fn update_power_config(config: &PowerConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO PowerConfig (
                is_power_saving, on_battery_shutdown_after_secs, adc_multiplier_override, wait_bluetooth_secs, 
                sds_secs, ls_secs, min_wake_secs, device_battery_ina_address
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8
            )",
            rusqlite::params![
                config.is_power_saving,
                config.on_battery_shutdown_after_secs,
                config.adc_multiplier_override,
                config.wait_bluetooth_secs,
                config.sds_secs,
                config.ls_secs,
                config.min_wake_secs,
                config.device_battery_ina_address
            ]
        )?;
        debug!("Saved PowerConfig to database");
        Ok(true)
    }

    pub fn update_network_config(config: &NetworkConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO NetworkConfig (
                wifi_enabled, wifi_ssid, wifi_psk, ntp_server, eth_enabled, address_mode,
                ip, gateway, subnet, dns, rsyslog_server
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11
            )",
            rusqlite::params![
                config.wifi_enabled,
                config.wifi_ssid,
                config.wifi_psk,
                config.ntp_server,
                config.eth_enabled,
                config.address_mode,
                config.ipv4_config.as_ref().map(|ipv4_config| ipv4_config.ip),
                config.ipv4_config.as_ref().map(|ipv4_config| ipv4_config.gateway),
                config.ipv4_config.as_ref().map(|ipv4_config| ipv4_config.subnet),
                config.ipv4_config.as_ref().map(|ipv4_config| ipv4_config.dns),
                config.rsyslog_server
            ]
        )?;
        debug!("Saved NetworkConfig to database");
        Ok(true)
    }

    pub fn update_display_config(config: &DisplayConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO DisplayConfig (
                screen_on_secs, gps_format, auto_screen_carousel_secs, compass_north_top, 
                flip_screen, units, oled, displaymode, heading_bold, wake_on_tap_or_motion
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10
            )",
            rusqlite::params![
                config.screen_on_secs,
                config.gps_format,
                config.auto_screen_carousel_secs,
                config.compass_north_top,
                config.flip_screen,
                config.units,
                config.oled,
                config.displaymode,
                config.heading_bold,
                config.wake_on_tap_or_motion
            ]
        )?;
        debug!("Saved DisplayConfig to database");
        Ok(true)
    }

    pub fn update_lora_config(config: &LoRaConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO LoRaConfig (
                use_preset, modem_preset, bandwidth, spread_factor, coding_rate, 
                frequency_offset, region, hop_limit, tx_enabled, tx_power, 
                channel_num, override_duty_cycle, sx126x_rx_boosted_gain, 
                override_frequency, ignore_incoming, ignore_mqtt
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15, ?16
            )",
            rusqlite::params![
                config.use_preset,
                config.modem_preset,
                config.bandwidth,
                config.spread_factor,
                config.coding_rate,
                config.frequency_offset,
                config.region,
                config.hop_limit,
                config.tx_enabled,
                config.tx_power,
                config.channel_num,
                config.override_duty_cycle,
                config.sx126x_rx_boosted_gain,
                config.override_frequency,
                // Rusqlite can't store a Vec<u32>, so we need to make it into Vec<u8>
                u32_vec_to_u8_vec(&config.ignore_incoming),
                config.ignore_mqtt
            ]
        )?;
        debug!("Saved LoRaConfig to database");
        Ok(true)
    }

    pub fn update_bluetooth_config(config: &BluetoothConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO BluetoothConfig (
                enabled, mode, fixed_pin
            ) VALUES (
                ?1, ?2, ?3
            )",
            rusqlite::params![
                config.enabled,
                config.mode,
                config.fixed_pin
            ]
        )?;
        debug!("Saved BluetoothConfig to database");
        Ok(true)
    }

    pub fn update_mqtt_config(config: &MqttConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO MqttConfig (
                enabled, address, username, password, encryption_enabled, json_enabled, 
                tls_enabled, root, proxy_to_client_enabled, map_reporting_enabled, 
                publish_interval_secs, position_precision
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12
            )",
            rusqlite::params![
                config.enabled,
                config.address,
                config.username,
                config.password,
                config.encryption_enabled,
                config.json_enabled,
                config.tls_enabled,
                config.root,
                config.proxy_to_client_enabled,
                config.map_reporting_enabled,
                config.map_report_settings.as_ref().map(|map_report_settings| map_report_settings.publish_interval_secs),
                config.map_report_settings.as_ref().map(|map_report_settings| map_report_settings.position_precision)
            ]
        )?;
        debug!("Saved MqttConfig to database");
        Ok(true)
    }

    pub fn update_serial_config(config: &SerialConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO SerialConfig (
                enabled, echo, rxd, txd, baud, timeout, mode, override_console_serial_port
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9
            )",
            rusqlite::params![
                config.enabled,
                config.echo,
                config.rxd,
                config.txd,
                config.baud,
                config.timeout,
                config.mode,
                config.override_console_serial_port
            ]
        )?;
        debug!("Saved SerialConfig to database");
        Ok(true)
    }

    pub fn update_externalnotification_config(config: &ExternalNotificationConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO ExternalNotificationConfig (
                enabled, output_ms, output, output_vibra, output_buzzer, active, alert_message, 
                alert_message_vibra, alert_message_buzzer, alert_bell, alert_bell_vibra, alert_bell_buzzer,
                use_pwm, nag_timeout, use_i2c_as_buzzer
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15
            )",
            rusqlite::params![
                config.enabled,
                config.output_ms,
                config.output_vibra,
                config.output_buzzer,
                config.active,
                config.alert_message,
                config.alert_message_vibra,
                config.alert_message_buzzer,
                config.alert_bell,
                config.alert_bell_vibra,
                config.alert_bell_buzzer,
                config.use_pwm,
                config.nag_timeout,
                config.use_i2s_as_buzzer
            ]
        )?;
        debug!("Saved ExternalNotificationConfig to database");
        Ok(true)
    }

    pub fn update_storeforward_config(config: &StoreForwardConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO StoreForwardConfig (
                enabled, heartbeat, records, history_return_max, history_return_window
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5
            )",
            rusqlite::params![
                config.enabled,
                config.heartbeat,
                config.records,
                config.history_return_max,
                config.history_return_window
            ]
        )?;
        debug!("Saved StoreForwardConfig to database");
        Ok(true)
    }

    pub fn update_rangetest_config(config: &RangeTestConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO RangeTestConfig (
                enabled, sender, save
            ) VALUES (
                ?1, ?2, ?3
            )",
            rusqlite::params![
                config.enabled,
                config.sender,
                config.save
            ]
        )?;
        debug!("Saved RangeTestConfig to database");
        Ok(true)
    }

    pub fn update_telemetry_config(config: &TelemetryConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO TelemetryConfig (
                device_update_interval, environment_update_interval, environment_measurement_enabled,
                environment_screen_enabled, environment_display_fahrenheit, air_quality_enabled,
                air_quality_interval, power_measurement_enabled, power_update_interval
                power_screen_enabled
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10
            )",
            rusqlite::params![
                config.device_update_interval,
                config.environment_update_interval,
                config.environment_measurement_enabled,
                config.environment_screen_enabled,
                config.environment_display_fahrenheit,
                config.air_quality_enabled,
                config.air_quality_interval,
                config.power_measurement_enabled,
                config.power_update_interval,
                config.power_screen_enabled
            ]
        )?;
        debug!("Saved TelemetryConfig to database");
        Ok(true)
    }

    pub fn update_cannedmessage_config(config: &CannedMessageConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO CannedMessageConfig (
                rotary1_enabled, inputbroker_pin_a, inputbroker_pin_b, inputbroker_pin_press,
                inputbroker_event_cw, inputbroker_event_ccw, inputbroker_event_press,
                updown1_enabled, enabled, allow_input_source, send_bell
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11
            )",
            rusqlite::params![
                config.rotary1_enabled,
                config.inputbroker_pin_a,
                config.inputbroker_pin_b,
                config.inputbroker_pin_press,
                config.inputbroker_event_cw,
                config.inputbroker_event_ccw,
                config.inputbroker_event_press,
                config.updown1_enabled,
                config.enabled,
                config.allow_input_source,
                config.send_bell
            ]
        )?;
        debug!("Saved CannedMessageConfig to database");
        Ok(true)
    }

    pub fn update_audio_config(config: &AudioConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO AudioConfig (
                codec2_enabled, ptt_pin, bitrate, i2s_ws, i2s_sd, i2s_din, i2c_sck
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7
            )",
            rusqlite::params![
                config.codec2_enabled,
                config.ptt_pin,
                config.bitrate,
                config.i2s_ws,
                config.i2s_sd,
                config.i2s_din,
                config.i2s_sck
            ]
        )?;
        debug!("Saved AudioConfig to database");
        Ok(true)
    }

    pub fn update_remotehardware_config(config: &RemoteHardwareConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        let serialized_pins = bincode::serialize(&config.available_pins).unwrap_or_default();
        db.conn.execute(
            "INSERT INTO RemoteHardwareConfig (
                enabled, allow_undefined_pin_access, available_pins
            ) VALUES (
                ?1, ?2, ?3
            )",
            rusqlite::params![
                config.enabled,
                config.allow_undefined_pin_access,
                serialized_pins
            ]
        )?;
        debug!("Saved RemoteHardwareConfig to database");
        Ok(true)
    }

    pub fn update_neighbourinfo_config(config: &NeighborInfoConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO NeighborInfoConfig (
                enabled, update_interval
            ) VALUES (
                ?1, ?2
            )",
            rusqlite::params![
                config.enabled,
                config.update_interval
            ]
        )?;
        debug!("Saved NeighborInfoConfig to database");
        Ok(true)
    }

    pub fn update_ambientlighting_config(config: &AmbientLightingConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO AmbientLightingConfig (
                led_state, led_current, red, green, blue
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5
            )",
            rusqlite::params![
                config.led_state,
                config.current,
                config.red,
                config.green,
                config.blue
            ]
        )?;
        debug!("Saved AmbientLightingConfig to database");
        Ok(true)
    }

    pub fn update_detectionsensor_config(config: &DetectionSensorConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO DetectionSensorConfig (
                enabled, minimum_broadcast_secs, state_broadcast_secs, send_bell,
                name, monitor_pin, detection_triggered_high, use_pullup

            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8
            )",
            rusqlite::params![
                config.enabled,
                config.minimum_broadcast_secs,
                config.state_broadcast_secs,
                config.send_bell,
                config.name,
                config.monitor_pin,
                config.detection_triggered_high,
                config.use_pullup
            ]
        )?;
        debug!("Saved DetectionSensorConfig to database");
        Ok(true)
    }

    pub fn update_paxcounter_config(config: &PaxcounterConfig) -> Result<bool> {
        let db = Database::open().expect("Failure to open database");
        db.conn.execute(
            "INSERT INTO PaxcounterConfig (
                enabled, paxcounter_update_interval
            ) VALUES (
                ?1, ?2
            )",
            rusqlite::params![
                config.enabled,
                config.paxcounter_update_interval
            ]
        )?;
        debug!("Saved PaxcounterConfig to database");
        Ok(true)
    }

    /// adds a new entry to the historical database for a NodeInfo Packet.
    pub fn add_gps_history_nodeinfo(node_info: &NodeInfo) -> Result<bool> {
        // Open DB
        let db = Database::open().expect("Failure to open database");

        // Add GPS History
        db.conn.execute(
            "INSERT INTO NodeGPSHistory (
                num, latitude_i, longitude_i, altitude, location_source, gps_timestamp, 
                altitude_source, pdop, hdop, vdop, gps_accuracy, ground_speed, 
                ground_track, fix_quality, fix_type, sats_in_view, sensor_id, 
                next_update, seq_number, precision_bits
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, 
                ?15, ?16, ?17, ?18, ?19
            )",
            rusqlite::params![
                node_info.num,
                node_info.position.as_ref().map(|position| position.latitude_i),
                node_info.position.as_ref().map(|position| position.longitude_i),
                node_info.position.as_ref().map(|position| position.altitude),
                node_info.position.as_ref().map(|position| position.location_source),
                node_info.position.as_ref().map(|position| position.timestamp),
                node_info.position.as_ref().map(|position| position.altitude_source),
                node_info.position.as_ref().map(|position| position.pdop),
                node_info.position.as_ref().map(|position| position.hdop),
                node_info.position.as_ref().map(|position| position.vdop),
                node_info.position.as_ref().map(|position| position.gps_accuracy),
                node_info.position.as_ref().map(|position| position.ground_speed),
                node_info.position.as_ref().map(|position| position.ground_track),
                node_info.position.as_ref().map(|position| position.fix_quality),
                node_info.position.as_ref().map(|position| position.fix_type),
                node_info.position.as_ref().map(|position| position.sats_in_view),
                node_info.position.as_ref().map(|position| position.sensor_id),
                node_info.position.as_ref().map(|position| position.next_update),
                node_info.position.as_ref().map(|position| position.seq_number),
                node_info.position.as_ref().map(|position| position.precision_bits)
            ]
        )?;


        debug!("Added more gps history for {}", node_info.num);
        Ok(true)
    }

    /// Adds a new radio history entry derived from a nodeinfo packet.
    pub fn add_radio_history_nodeinfo(node_info: &NodeInfo) -> Result<bool> {
        // Open DB
        let db = Database::open().expect("Failure to open database");

        // Add GPS History
        db.conn.execute(
            "INSERT INTO NodeRadioHistory (
                num, snr, channel_utilization, air_util_tx, channel, via_mqtt, hops_away
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7
            )",
            rusqlite::params![
                node_info.num,
                node_info.snr,
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.channel_utilization),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.air_util_tx),
                node_info.channel,
                node_info.via_mqtt,
                node_info.hops_away
            ]
        )?;


        debug!("Added more radio history for {}", node_info.num);
        Ok(true)
    }

    /// Adds a new radio history entry derived from a nodeinfo packet.
    pub fn add_hardware_history_nodeinfo(node_info: &NodeInfo) -> Result<bool> {

        // Open DB
        let db = Database::open().expect("Failure to open database");

        // Add GPS History
        db.conn.execute(
            "INSERT INTO NodeHardwareHistory (
                num, user_lname, user_sname, hw_model, is_licensed, role, battery_level, voltage
            ) VALUES (
                ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8
            )",
            rusqlite::params![
                node_info.num,
                node_info.user.as_ref().map(|user| user.long_name.clone()),
                node_info.user.as_ref().map(|user| user.short_name.clone()),
                node_info.user.as_ref().map(|user| user.hw_model),
                node_info.user.as_ref().map(|user| user.is_licensed),
                node_info.user.as_ref().map(|user| user.role),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.battery_level),
                node_info.device_metrics.as_ref().map(|device_metrics| device_metrics.voltage),
            ]
        )?;

        debug!("Added more hardware history for {}", node_info.num);
        Ok(true)
    }

    /// Returns a list of all nodes and their current info.
    pub fn read_node_list() -> Result<Vec<NodeInfo>> {
        let db = Database::open().expect("Failure to open database for reading");

        let mut stmt = db.conn.prepare("SELECT num, user_id, user_lname, user_sname, 
                                        hw_model, is_licensed, role, battery_level, 
                                        voltage, latitude_i, longitude_i, altitude, 
                                        location_source, gps_timestamp, altitude_source, 
                                        pdop, hdop, vdop, gps_accuracy, ground_speed, 
                                        ground_track, fix_quality, fix_type, sats_in_view, 
                                        sensor_id, next_update, seq_number, precision_bits, 
                                        snr, last_heard, channel_utilization, air_util_tx, 
                                        channel, via_mqtt, hops_away FROM Node")?;

        // For each node in the query, map to NodeInfo struct
        let node_iter = stmt.query_map([], |row| {
            Ok(NodeInfo {
                num: row.get(0)?,
                user: {
                    Some(User {
                        id: row.get(1).unwrap_or("NULL".to_owned()),
                        long_name: row.get(2).unwrap_or("NULL".to_owned()),
                        short_name: row.get(3).unwrap_or("NULL".to_owned()),
                        macaddr: vec![],
                        hw_model: row.get(4).unwrap_or_default(),
                        is_licensed: row.get(5).unwrap_or(false),
                        role: row.get(6).unwrap_or_default(),
                    })
                },
                position: {
                    Some(Position {
                        latitude_i: row.get(9).unwrap_or(0),
                        longitude_i: row.get(10).unwrap_or(0),
                        altitude: row.get(11).unwrap_or(0),
                        time: 0,
                        location_source: row.get(12).unwrap_or_default(),
                        altitude_source: row.get(14).unwrap_or_default(),
                        timestamp: row.get(13).unwrap_or(0),
                        timestamp_millis_adjust: 0,
                        altitude_hae: 0,
                        altitude_geoidal_separation: 0,
                        pdop: row.get(15).unwrap_or(0),
                        hdop: row.get(16).unwrap_or(0),
                        vdop: row.get(17).unwrap_or(0),
                        gps_accuracy: row.get(18).unwrap_or(0),
                        ground_speed: row.get(19).unwrap_or(0),
                        ground_track: row.get(20).unwrap_or(0),
                        fix_quality: row.get(21).unwrap_or(0),
                        fix_type: row.get(22).unwrap_or(0),
                        sats_in_view: row.get(23).unwrap_or(0),
                        sensor_id: row.get(24).unwrap_or(0),
                        next_update: row.get(25).unwrap_or(0),
                        seq_number: row.get(26).unwrap_or(0),
                        precision_bits: row.get(27).unwrap_or(0),
                    })
                },
                snr: row.get(28)?,
                last_heard: row.get(29)?,
                device_metrics: {
                    Some(DeviceMetrics {
                        battery_level: row.get(7).unwrap_or(101),
                        voltage: row.get(8).unwrap_or(0.0),
                        channel_utilization: row.get(30).unwrap_or(0.0),
                        air_util_tx: row.get(31).unwrap_or(0.0),
                    })
                },
                channel: row.get(32)?,
                via_mqtt: row.get(33)?,
                hops_away: row.get(34)?,
            })
        })?;

        let nodes: Result<Vec<NodeInfo>> = node_iter.collect();
        //debug!("Nodes from DB: {:?}", nodes);
        nodes
    }

}