use std::{env, process};

use figment::{Figment, providers::{Format, Toml, Env}};

use lazy_static::lazy_static;
use meshtastic::api::{StreamApi, StreamHandle};
use meshtastic::utils;
use meshtastic::protobufs::*;

use packets::parse;
use tokio::{net::TcpStream, sync::mpsc::UnboundedReceiver};
use tracing::{debug, error, info, Level};
use tracing_subscriber::FmtSubscriber;

mod nodes;
mod packets;
mod db;
mod config;
mod gui;
use crate::config::AppConfig;

// Setup the config globally because I can't figure out how to pass it to functions I don't call directly.
// This is evaluated at runtime, and not compilation. \o/
// Also, we try to merge configs. so envvars > config > defaults > error
lazy_static! {
    pub static ref CONFIG: AppConfig = {
        let config = Figment::new()
            .merge(Toml::file("config.toml"))
            .merge(Env::prefixed("yamm_"))
            .extract()
            .unwrap_or_default();
        tracing::info!("Loaded Config!");
        tracing::debug!("full config: {:?}", config);
        config
    };
}

async fn run_server() {
    // set up new connection
    let stream_api = StreamApi::new();

    // Different connection types have different setups.
    let mut tcp_stream: StreamHandle<TcpStream>;
    //let mut serial_stream: StreamHandle<SerialStream>; // SerialStream isn't a valid type???
    match &CONFIG.meshconfig.connection_type {
        config::ConnectionType::Tcp => {
            // Build the tcp stream object from config.
            let address = &CONFIG.meshconfig.ip;
            let port = &CONFIG.meshconfig.port.to_string();
            let tcp = format!("{}:{}", address, port);

            // Connect over TCP
            let tcp_stream = utils::stream::build_tcp_stream(tcp).await.expect("Invalid TCP Stream");
            let (mut decoded_listener, stream_api) = stream_api.connect(tcp_stream).await;
            let config_id = utils::generate_rand_id();
            let stream_api = stream_api.configure(config_id).await.expect("Invalid Stream Config");

            // Read loop, exit with Ctrl+C or disconnecting node.
            while let Some(decoded) = decoded_listener.recv().await {
                // New Message!
                debug!("Received: {:?}", decoded);
                // Parse message
                parse(&decoded);
            }
        
            // Note that in this specific example, this will only be called when
            // the radio is disconnected, as the above loop will never exit.
            // Typically you would allow the user to manually kill the loop,
            // for example with tokio::select!.
            let _stream_api = stream_api.disconnect().await.expect("Failed to disconnect stream");
        },
        config::ConnectionType::Serial => {
            let available_ports = utils::stream::available_serial_ports().expect("Couldn't find serial ports");
            info!("Available ports, please set in config.: {:?}", available_ports);
            todo!("Not yet Implemented. See https://git.volkor.me/Volkor/yamm/-/issues/2")
        },
        config::ConnectionType::Mqtt => {
            todo!("Not yet Implemented. See https://git.volkor.me/Volkor/yamm/-/issues/3")
        },
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    // Initialise Logging - this isn't final.
    let subscriber = FmtSubscriber::builder()
        // all events higher than debug are shown
        .with_max_level(Level::DEBUG)
        .finish();
    tracing::subscriber::set_global_default(subscriber)
        .expect("setting default subscriber failed");

    // Hello World!
    info!("Y(et) A(nother) M(eshtastic) S(erver) - v0.1.0");


    // Do we run a server or gui?
    let args: Vec<String> = env::args().collect();

    // display help on no args, exit.
    if args.len() < 2 {
        info!("Usage: yams <mode>");
        info!("Modes: 'server', 'gui'");
        process::exit(1);
    }

    // match on gui or server
    match args[1].as_str() {
        "server" => run_server().await,
        "gui" => gui::run_gui().await,
        _ => {
            error!("Unknown mode: {}", args[1]);
            info!("Usage: yams <mode>");
            info!("Modes: 'server', 'gui'");
            process::exit(1);
        }
    }

    Ok(())
}