use std::path::Path;

use config::device_config;
use from_radio::PayloadVariant;
use meshtastic::protobufs::*;
use tracing::{debug, error};

use crate::db::Database;

/// Parses all incoming packets, handing them out to the correct handler.
pub fn parse(message: &FromRadio) -> bool {
    // debug!("Parsing Message from radio: {:?}", &message);

    // Handle all types of messages from the radio here.
    match message.payload_variant {
        // A normal message
        Some(PayloadVariant::Packet(ref packet)) => {
            debug!("New Message Packet!");
            parse_packet(packet);
        },
        // Information about the local node
        Some(PayloadVariant::MyInfo(ref info)) => {
            debug!("New MyNodeInfo Packet");
            parse_myinfo(info);
        },
        // Information about a remote node
        Some(PayloadVariant::NodeInfo(ref node_info)) => {
            debug!("New NodeInfo Packet!");
            parse_nodeinfo(node_info);
        },
        // Information about the local node's config
        Some(PayloadVariant::Config(ref config)) => {
            debug!("New Config Packet!");
            parse_config(config);
        },
        // Information about the local node's module config
        Some(PayloadVariant::ModuleConfig(ref module_config)) => {
            debug!("New ModuleConfig Packet!");
            parse_module_config(module_config);
        },
        // Information about one of the 8 channels in (potential) use.
        Some(PayloadVariant::Channel(ref channel_info)) => {
            debug!("New Channel Packet!");
            parse_channel_info(channel_info);
        },
        // Information about the queue (messages to send soon)
        Some(PayloadVariant::QueueStatus(ref queue_status)) => {
            debug!("New Message Packet!");
            parse_queue_info(queue_status);
        },

        // Unhandled or empty packet
        _ => {
            error!("Unhandled Packet: {:?}", message);
        }
    }

    // Return success
    true
}

fn parse_packet(packet: &MeshPacket) {
    // Handle packet data
    error!("Unhandled MeshPacket Packet: {:#?}", &packet);
    
}

fn parse_myinfo(packet: &MyNodeInfo) {
    debug!("Parsing MyNodeInfo Packet: {:#?}", &packet);
    let _ = Database::update_mynodeinfo(packet);
}

fn parse_nodeinfo(node_info: &NodeInfo) {
    // Handle NodeInfo data
    debug!("Parsing NodeInfo Packet: {:?}", &node_info);
    let _ = Database::update_nodeinfo(node_info);
    // Save History for this nodeinfo packet
    let _ = Database::add_gps_history_nodeinfo(node_info);
    let _ = Database::add_radio_history_nodeinfo(node_info);
    let _ = Database::add_hardware_history_nodeinfo(node_info);
}

fn parse_config(config: &Config) {
    // Handle Config data
    debug!("Parsing Config Packet: {:#?}", &config);
    
    // Match and destructure the payload_variant
    if let Some(payload_variant) = config.payload_variant.clone() {
        match payload_variant {
            config::PayloadVariant::Device(device_config) => {
                let _ = Database::update_deviceconfig(&device_config);
            },
            config::PayloadVariant::Position(position_config) => {
                let _ = Database::update_position_config(&position_config);
            },
            config::PayloadVariant::Power(power_config) => {
                let _ = Database::update_power_config(&power_config);
            },
            config::PayloadVariant::Network(network_config) => {
                let _ = Database::update_network_config(&network_config);
            },
            config::PayloadVariant::Display(display_config) => {
                let _ = Database::update_display_config(&display_config);
            },
            config::PayloadVariant::Lora(lora_config) => {
                let _ = Database::update_lora_config(&lora_config);
            },
            config::PayloadVariant::Bluetooth(bluetooth_config) => {
                let _ = Database::update_bluetooth_config(&bluetooth_config);
            },
        }
    }
}

fn parse_module_config(module_config: &ModuleConfig) {
    debug!("Parsing Module Packet: {:#?}", &module_config);
    // Handle ModuleConfig data
    // Match and destructure the payload_variant
    if let Some(payload_variant) = module_config.payload_variant.clone() {
        match payload_variant {
            module_config::PayloadVariant::Mqtt(mqtt_config) => {
                let _ = Database::update_mqtt_config(&mqtt_config);
            },
            module_config::PayloadVariant::Serial(serial_config) => {
                let _ = Database::update_serial_config(&serial_config);
            },
            module_config::PayloadVariant::ExternalNotification(external_notification_config) => {
                let _ = Database::update_externalnotification_config(&external_notification_config);
            },
            module_config::PayloadVariant::StoreForward(store_forward_config) => {
                let _ = Database::update_storeforward_config(&store_forward_config);
            },
            module_config::PayloadVariant::RangeTest(range_test_config) => {
                let _ = Database::update_rangetest_config(&range_test_config);
            },
            module_config::PayloadVariant::Telemetry(telemetry_config) => {
                let _ = Database::update_telemetry_config(&telemetry_config);
            },
            module_config::PayloadVariant::CannedMessage(canned_message_config) => {
                let _ = Database::update_cannedmessage_config(&canned_message_config);
            },
            module_config::PayloadVariant::Audio(audio_config) => {
                let _ = Database::update_audio_config(&audio_config);
            },
            module_config::PayloadVariant::RemoteHardware(remote_hardware_config) => {
                let _ = Database::update_remotehardware_config(&remote_hardware_config);
            },
            module_config::PayloadVariant::NeighborInfo(neighbor_info_config) => {
                let _ = Database::update_neighbourinfo_config(&neighbor_info_config);
            },
            module_config::PayloadVariant::AmbientLighting(ambient_lighting_config) => {
                let _ = Database::update_ambientlighting_config(&ambient_lighting_config);
            },
            module_config::PayloadVariant::DetectionSensor(detection_sensor_config) => {
                let _ = Database::update_detectionsensor_config(&detection_sensor_config);
            },
            module_config::PayloadVariant::Paxcounter(paxcounter_config) => {
                let _ = Database::update_paxcounter_config(&paxcounter_config);
            },
        }
    }
}

fn parse_channel_info(channel_info: &Channel) {
    debug!("Parsing Channel Packet: {:#?}", &channel_info);
    // Handle ChannelInfo data
    let _ = Database::update_channel(channel_info);
}

fn parse_queue_info(queue_info: &QueueStatus) {
    error!("Unhandled Queue Packet: {:#?}", &queue_info);
    // Handle QueueInfo data
}